/* 
   jplay2 - JACK audio player for multimedia files

   Copyright (C) 2007 Robin Gareus

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "player.h"
#include "buffer.h"
#include "xmalloc.h"

extern int m_channels;
extern int m_samplerate;
extern int want_verbose;
extern int want_quiet;

//////  buffer ///////


jack_nframes_t fillBuffer_cache(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames, 
				jack_nframes_t m_position, audiocache *cache, int *fadeedges )
{
	int i;
	int have_data = 0;
	jack_nframes_t written = 0;
	audiocache *cptr = cache;
	while (cptr && written < frames) {
		if ((cptr->state&1)!=1 ) { cptr=cptr->next; continue; }
		if (cptr->start <= (m_position + written) && cptr->end > (m_position+ written)) { 
//			printf("found one : (%i) %i - %i\n",(m_position+written),cptr->start,cptr->end);
			have_data = 1;
			jack_nframes_t len;
			jack_nframes_t off;
			len = min (frames-written, cptr->end-(m_position + written));
			off =  (m_position + written)-cptr->start;
//			printf("copy:  off:%i - len:%i\n",off,len);
			for (i=0;i<m_channels;i++) {
				int j,k;
      				for(k=written,j=off; j < off+len; j++, k++)
					bufferptrs[i][k]= cptr->d[i][j];
			}
			written += len;
			cptr = cache; continue; // restart cache search at new start pos.
		}
		cptr=cptr->next; 
	}

	if (*fadeedges < 0 ) return (written);

	// TODO if prev frame was padded - fade in .
	if (*fadeedges) {
		*fadeedges=0;
		int f; 
		for (f=0; f< min(written,FADELEN); f++)
			for (i=0;i<m_channels;i++) 
				bufferptrs[i][f]*=(f/FADELEN);
	}

	if (written < frames) {
		if (want_verbose)
			printf("  padding %i zeros @%i (loss: %2i%%)\n", (frames-written),m_position,(frames-written)*100/frames);
		*fadeedges=1;
		for (i=0;i<m_channels;i++) 
			memset(bufferptrs[i]+written,0, sizeof (jack_default_audio_sample_t) * (frames-written));
	}

	if (0 && *fadeedges) {
		int f; 
		for (f=0; f< min(written,FADELEN); f++)
			for (i=0;i<m_channels;i++) 
				bufferptrs[i][(written-f)]*=(f/FADELEN);
	}

	return (frames);
}

audiocache *newcl(audiocache *cache) {
	int i;
	audiocache *cl;

	cl=xcalloc(1,sizeof(audiocache));
	cl->next= NULL;
	cl->state=0;
//	cl->d = xcalloc(CACHE_S*m_channels,sizeof(jack_default_audio_sample_t));
	cl->d = xcalloc(m_channels,sizeof(jack_default_audio_sample_t*));
	for (i=0;i<m_channels;i++) 
		cl->d[i] = xcalloc(CACHE_S,sizeof(jack_default_audio_sample_t));

	// append node
	cl->state&=~1;
	audiocache *cptr = cache;
	while (cptr && cptr->next) cptr=cptr->next;
	if (cptr) cptr->next=cl;

	return (cl);
}

audiocache *getcl( audiocache *cache ) {
	audiocache *cptr = cache;
	while (cptr) {
		if (cptr->state == 0 ) return cptr;
		cptr=cptr->next;
	}
	return (newcl(cache));
}

void freecl(jack_nframes_t m_position, audiocache *cache ) {
	audiocache *cptr = cache;
	while (cptr) {
		if ((cptr->state&1)!=1 ) { cptr=cptr->next; continue; }
		if ((cptr->end + 5*m_samplerate ) <  m_position ) { // XXX : adjust total cache size 
			cptr->state&=~1;
		}
		cptr=cptr->next;
	}
}

void clearcache(audiocache *cache) {
	audiocache *cptr = cache;
	while (cptr) {
		cptr->state=0;
		cptr=cptr->next;
	}
}

int testcls(jack_nframes_t p, audiocache *cache) {
	audiocache *cptr = cache;
	while (cptr) {
		if ((cptr->state&1)!=1 ) { cptr=cptr->next; continue; }
		if (cptr->start <= p && cptr->end > p) { 
			return (1);
		}
		cptr=cptr->next;
	}
	return (0);
}

