/* 
 *  Copyright (C) 2006 Robin Gareus  <robin AT gareus.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <jack/jack.h>
#include <jack/transport.h>
#include <math.h>
#include <string.h>

int j_scrubpos = 0;
int j_scrubmax = 1;

jack_client_t *j_client = NULL;
jack_port_t **j_output_port; 
jack_default_audio_sample_t **j_output_bufferptrs; // 

jack_nframes_t j_bufsiz = 64;
jack_nframes_t j_srate = 48000;
jack_nframes_t j_latency = 0;


// extern prototypes //

jack_nframes_t fillBuffer( jack_default_audio_sample_t ** bufferptrs, jack_nframes_t frames );
void sampleseek (jack_nframes_t sample);
void jadio_realloc(jack_nframes_t frames );

extern jack_nframes_t m_length; // length of file in frames.
extern jack_nframes_t m_resampled_length; // length of file in jack-frames.
extern int m_channels ;    // channels of the file
extern int m_samplerate;    // sample rate of the file
extern float m_fResampleRatio;
extern int p_scrublen;
extern int c_run;
extern int want_transport;
extern int want_scrub;
extern int want_verbose;
extern int want_quiet;

extern float want_speed;

#define SLP 3 // sync low-pass buffer size

/* jadio - jack */
jack_nframes_t playhead = 48000*0; // TODO set to EOF when playing reverse
jack_nframes_t synchead[SLP]; 

void jackreadAudio( jack_default_audio_sample_t **out, jack_transport_state_t ts, jack_nframes_t position, jack_nframes_t nframes )
{
	int i;
	jack_nframes_t scrub_pos=0;

	// adjust resample_length to match current playback speed
	if (want_speed == 0.0)  {
		for (i=0;i<m_channels;i++) {
			int ii;
			for (ii=0;ii<(nframes>>1);ii++) { 
				out[i][ii]= (jack_default_audio_sample_t) 0;
			}
		}
	}

	if (want_transport)  {
#if 0 // relative transport
		scrub_pos=playhead;
		for (i=SLP-1;i>0;i--) synchead[i]=synchead[i-1];
		synchead[0] = position;
		float delta = 0.0;
		float sum = 0.0;
		for (i=0;i<SLP-1;i++) {
			//printf ("XXX %i\n",synchead[i+1], synchead[i]);
			delta += ((float)synchead[i]-(float)synchead[i+1]) / ((float)(i+1));
			sum += 1.0/(i+1);
		}
		delta/=sum;
		printf ("%.3f %.3f\n",delta,delta/(float)nframes);

		// TODO:
		// fit playspeed - SLP data 
		// extrapolate no. of frames.
		
		if (0 && delta!=0.0) {
  want_speed = (float)nframes/delta;
  if (want_speed < -20.0 ) want_speed =-20.0;
  if (want_speed >  20.0 ) want_speed = 20.0;
  m_fResampleRatio = (float) j_srate / (float) m_samplerate * fabs(want_speed);
  m_resampled_length= (jack_nframes_t) ceil((float)m_length * m_fResampleRatio);
  //jadio_realloc( j_bufsiz );
  }
		// set playhead to extrapolated pos
		if (ts == JackTransportRolling)
			playhead+=(nframes / m_fResampleRatio);
#else
		scrub_pos=position / m_fResampleRatio; 
		playhead=position;
#endif
	}else  {
		scrub_pos=playhead;
		if (want_speed < 0) 
			playhead-= (nframes / m_fResampleRatio); // FIXME: increment _after_ loop or only if running?? - speed
		else 	playhead+= (nframes / m_fResampleRatio); // FIXME: increment _after_ loop or only if running?? - speed
	}
	if (want_transport && want_scrub)  {
		if (want_scrub && ts == JackTransportStopped && j_scrubmax>1) {
			scrub_pos+=(j_scrubpos) * j_bufsiz;
			j_scrubpos=(j_scrubpos+1)%(j_scrubmax);
		} else j_scrubpos=0;
	} else  {
		j_scrubpos=0;
	}

	// compensate output port latency
#if 1
	if (scrub_pos > j_latency) scrub_pos-=j_latency;
	else scrub_pos = 0;
#endif

	if (want_speed < 0 && scrub_pos > j_bufsiz) scrub_pos-=j_bufsiz;
	else if (scrub_pos < 0) scrub_pos=0;

	if ( !want_transport  && (scrub_pos >= m_resampled_length )) {
		if(want_verbose) fprintf(stderr, "End Of File. (%lu/%lu)\n",(unsigned long int)scrub_pos,(unsigned long int)m_resampled_length);
		c_run=0; // stop playback on x-run.
	}

	if ( (scrub_pos < 0 )  || (scrub_pos >= m_resampled_length ) 
		|| (want_transport && (!want_scrub && ts != JackTransportRolling))
	   ){ // if transport is not rolling - or file ended : remain silent.

		for (i=0;i<m_channels;i++) 
			memset(out[i],0, sizeof (jack_default_audio_sample_t) * nframes);

		if (ts == JackTransportStarting && !((scrub_pos < 0 )||(scrub_pos >= m_resampled_length)) ) {
			sampleseek(scrub_pos); 
		}
		return;
	}
	sampleseek(scrub_pos); // set absolute audio sample position

	jack_nframes_t rv = fillBuffer(out, nframes);

	if (want_speed < 0) {
		// reverse audio for varispeed.
		int ii;
		for (i=0;i<m_channels;i++) {
			for (ii=0;ii<(nframes>>1);ii++) { 
				jack_default_audio_sample_t tmp=out[i][ii];
				out[i][ii]=out[i][nframes-ii];
				out[i][nframes-ii]=tmp;
			}
		}
	}

	if (rv != nframes) {
		if (c_run)
			if(!want_quiet) fprintf(stderr, "Soundoutput buffer underrun: %i/%i (decoder too slow)\n",rv,nframes);
		if (want_transport) 
			c_run=0; // stop playback on x-run.
		for (i=0;i<m_channels;i++) // silence.
			memset(out[i],0, sizeof (jack_default_audio_sample_t) * nframes);
	}
}


/*
 * The process callback for this JACK application is called in a
 * special realtime thread once for each audio cycle.
 */
int jack_audio_callback (jack_nframes_t nframes, void *arg)
{
	jack_default_audio_sample_t **out = j_output_bufferptrs; // 
	jack_position_t	jack_position;
	jack_transport_state_t ts;
	int i;
	jack_nframes_t my_tot_latency = 0;

	for (i=0;i<m_channels;i++) {

		jack_nframes_t my_latency = jack_port_get_total_latency(j_client,j_output_port[i]);
		if (my_latency > my_tot_latency) my_tot_latency = my_latency;

		//TODO: per channel compensation!

//		printf("DEBUG: c=%i pl=%i tl=%i\n",i,jack_port_get_latency(j_output_port[i]) ,jack_port_get_total_latency(j_client,j_output_port[i]));

		out[i] = jack_port_get_buffer (j_output_port[i], nframes);
//		memset(out[i],0, sizeof (jack_default_audio_sample_t) * nframes);
	}
	if (my_tot_latency != j_latency) {
		j_latency = my_tot_latency;
		if(want_verbose) printf("latency compensation: %i frames \n",j_latency);
	}

	ts=jack_transport_query(j_client, &jack_position);

	jackreadAudio(out, ts, jack_position.frame, nframes);
#if 0
	static int leier =0;
	leier=(leier+1)%400;
	if (leier%20) {
		m_fResampleRatio = (float) j_srate / (float) m_samplerate * (1.0+(((float)leier-200.0)/2000.0));
		m_resampled_length= (jack_nframes_t) ceil((float)m_length * m_fResampleRatio);
	}
#endif
	return(0);
}

/* when jack shuts down... */
void jack_shutdown_callback(void *arg)
{
	jack_client_t *remember = j_client;
	j_client=NULL;
	if (!want_quiet)
		fprintf(stderr,"jack server shutdown.\n");
	jack_client_close (remember);
	if (!want_quiet)
		fprintf(stderr,"closed zombified jack server connection .\n");
}

/* when jack changes its samplerate ... */
int jack_srate_callback(jack_nframes_t nframes, void *arg)
{
	if(want_verbose) fprintf(stderr,"jack sample-rate: %i\n",nframes); 
	if(want_speed != 1.0) fprintf(stderr,"changing speed to %.2f\n",want_speed); 
	j_srate = nframes;
	if(want_speed == 0.0) return(0);
	m_fResampleRatio = (float) j_srate / (float) m_samplerate * fabs(want_speed);
	m_resampled_length= (jack_nframes_t) ceil((float)m_length * m_fResampleRatio);
	// FIXME: actually we should realloc the resampling buffer when this happens 
	// but this requires a non RT callback. -> buffersize change
	return(0); 
}

/* when jack changes its buffer size ... 
 * -  here we can do non realtime stuff // eg. realloc 
 */
int jack_bufsiz_callback(jack_nframes_t nframes, void *arg)
{
	if(want_verbose) fprintf(stderr,"jack buffer size: %i\n",nframes);
	j_bufsiz = nframes;
	jadio_realloc( j_bufsiz );
	j_scrubpos = 0;
	j_scrubmax = (j_bufsiz!=0)?(int)ceil((double) p_scrublen/j_bufsiz):1;
	if(!want_quiet && want_transport && want_scrub) fprintf(stderr,"scrublength: %i buffers (%.1f ms - %.1f Hz)\n",j_scrubmax,1000.0*(double)j_scrubmax*(double)j_bufsiz/(double)j_srate,(double)j_srate/(double)j_scrubmax/(double)j_bufsiz);
	return(0); 
}



void close_jack(void)
{
	if (j_client) {
		jack_deactivate(j_client);
		jack_client_close (j_client);
	}
	j_client=NULL;
}

int jack_connected() {
	if (j_client) return(1);
	return(0);
}

extern int jack_portc;
extern char **jack_portnames;

void connect_jackports()  {
	int i;
	if (!j_client) return;
	if (jack_portc ==0 ) return;

	int port_flags = JackPortIsInput;
	if (jack_portc<0) port_flags |= JackPortIsPhysical;

	const char **found_ports = jack_get_ports(j_client, NULL, NULL, port_flags);
	if (!found_ports) return;
	if (jack_portc<0) { // autoconnect
		int myc=0;
		for (i = 0; found_ports[i] && myc < m_channels; i++) {
			if (jack_connect(j_client, jack_port_name(j_output_port[myc]), found_ports[i])) {
				fprintf(stderr,"can not connect to jack output\n");
			} else if (want_verbose)
				printf("connected %s -> %s\n",jack_port_name(j_output_port[myc]), found_ports[i]);
			// TODO: connect mono samples to first two outputs only if (p_two_mono)
			if (m_channels!=1 || i!=0) myc++;
		}
	} else { //  -J <port-name>,..
		int myc=0;
		for (i = 0; i < jack_portc ; i++) {
			int j;
			for (j = 0; found_ports[j]; j++) {
				if (!strcasecmp(found_ports[j],jack_portnames[i])) {
					if (jack_connect(j_client, jack_port_name(j_output_port[myc]), found_ports[j])) {
						fprintf(stderr,"can not connect to jack output\n");
					} else if (want_verbose)
						printf("connected %s -> %s\n",jack_port_name(j_output_port[myc]), found_ports[j]);
					if (myc < m_channels-1) myc++;
					break;
				}
			}
		}
	}
}

void open_jack(void) 
{
	int i;
	if (j_client) {
		fprintf(stderr,"already connected to jack.\n");
		return;
	}

	for (i=0;i< SLP; i++) synchead[i]=0;

	i = 0;
	do {
		char jackid[16];
		snprintf(jackid,16,"jadio-%i",i);
		j_client = jack_client_open (jackid, JackUseExactName, NULL);
	} while (j_client == 0 && i++<16);

	if (!j_client) {
		fprintf(stderr, "could not connect to jack.\n");
		return;
	}	

	jack_on_shutdown (j_client, jack_shutdown_callback, NULL);
	jack_set_process_callback(j_client,jack_audio_callback,NULL);
	jack_set_sample_rate_callback (j_client, jack_srate_callback, NULL);
	jack_set_buffer_size_callback (j_client, jack_bufsiz_callback, NULL);

	j_output_port= calloc(m_channels,sizeof(jack_port_t*));
	j_output_bufferptrs = calloc(m_channels,sizeof(jack_default_audio_sample_t*));

	for (i=0;i<m_channels;i++) {
		char channelid[16];
		snprintf(channelid,16,"output-%i",i);
		j_output_port[i] = jack_port_register (j_client, channelid, JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
		if (!j_output_port[i]) {
			fprintf(stderr, "no more jack ports availabe.\n");
			close_jack();
			return;
		}
	}
	jack_srate_callback(jack_get_sample_rate(j_client),NULL); // init j_srate
	jack_bufsiz_callback(jack_get_buffer_size(j_client),NULL); // init j_bufsiz & alloc

	jack_activate(j_client);

	connect_jackports();
}

// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=--=-=-=-=

void dothejack(void)
{
	open_jack();  // try to connect to jack

	if (jack_connected()) {
		j_scrubpos = 0;
		j_scrubmax = (j_bufsiz!=0)?(int)ceil(p_scrublen/j_bufsiz):1;
	} 
}

int c_run = 1;
void loopthejack(void) {
	if (!jack_connected()) return;

	while (c_run) { 
		if (!jack_connected()) dothejack();
		sleep (1);
	}
}
