/* 
   jplay2 - JACK audio player for multimedia files

   Copyright (C) 2007 Robin Gareus

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "player.h"
#include "buffer.h"
#include "xmalloc.h"

extern int m_channels;
extern int m_samplerate;
extern int want_quiet;
extern int want_osc;
extern int want_verbose;
extern int c_run;
extern float want_speed;
extern jack_nframes_t j_srate;
extern float m_fResampleRatio;
extern jack_nframes_t m_length; // length of file in frames.
extern jack_nframes_t m_resampled_length; // length of file in jack-frames.
extern jack_nframes_t j_bufsiz;

void jadio_realloc(jack_nframes_t frames );

#ifdef HAVE_LIBLO

#include <lo/lo.h>
#include <fcntl.h>
#include <pthread.h>

int oscb_speed (const char *path, const char *types, lo_arg **argv, int argc, lo_message msg, void *user_data){
  if (want_verbose) printf("%s <- f:%f\n", path, argv[0]->f);
  want_speed = argv[0]->f;
  m_fResampleRatio = (float) j_srate / (float) m_samplerate * fabs(want_speed);
  m_resampled_length= (jack_nframes_t) ceil((float)m_length * m_fResampleRatio);
  jadio_realloc( j_bufsiz );
  return(0);
}

int oscb_quit (const char *path, const char *types, lo_arg **argv, int argc, lo_message msg, void *user_data){
  printf("OSC 'quit' command recv.\n");
  c_run=0;
  return(0);
}

static void oscb_error(int num, const char *m, const char *path) {
  fprintf(stderr, "liblo server error %d in path %s: %s\n", num, path, m);
}

lo_server_thread osc_server = NULL;

#define PORTAINC 1

int initialize_osc(void) {
  char tmp[255];
  int j;
  uint32_t port = (want_osc>100 && want_osc< 60000)?want_osc:7000;
  for (j=0; j < PORTAINC; ++j) {
    snprintf(tmp, sizeof(tmp), "%d", port);
    printf("OSC trying port:%i\n",port);
    if ((osc_server = lo_server_thread_new (tmp, oscb_error))) break;
    fprintf (stderr,"OSC port %i is in use.\n", port);
    port++;
  }

  if (osc_server) {
    char *urlstr;
    urlstr = lo_server_thread_get_url (osc_server);
    printf("OSC server name: %s\n",urlstr);
    free (urlstr);

    lo_server_thread_add_method(osc_server, "/jplay2/set_transport_speed", "f",  &oscb_speed, NULL);
    lo_server_thread_add_method(osc_server, "/jplay2/quit", "", &oscb_quit, NULL);

    lo_server_thread_start(osc_server);
    if(!want_quiet) printf("OSC server started on port %i\n",port);
    return (0);
  } 

  if(!want_quiet) printf("OSC start failed.");
  return(1);
}

void shutdown_osc(void) {
  if (!osc_server) return;
  lo_server_thread_stop(osc_server);
  if(!want_quiet) printf("OSC server shut down.\n");
}
#endif

/* vi:set ts=8 sts=2 sw=2: */
