/* 
   jplay2 - JACK audio player for multimedia files

   Copyright (C) 2007 Robin Gareus

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 
   as published by the Free Software Foundation;

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "xmalloc.h"
#include "player.h"
#include "decoder.h"
#include "buffer.h"


int m_filetype = SFT_NONE;
int m_samplerate = 48000;    // sample rate of the file
int m_channels = 2;    // channels of the file
jack_nframes_t m_length = 0; // length of file in frames. @ orig samplerate
jack_nframes_t m_resampled_length; // length of file in jack-frames. @ jack samplerate
float m_fResampleRatio = 1.0;


void openfile(char *filename) {
	//TODO: allow user to choose or autodetect format

	if (m_filetype == SFT_NONE)
		openfile_sf(filename);
//  if file ends with .mov: try libqt first
	if (m_filetype == SFT_NONE && strlen(filename)>4 && !strncasecmp(filename+(strlen(filename)-4),".mov",4) )
		openfile_qt(filename);
	if (m_filetype == SFT_NONE)
//		openfile_ff(filename);
		openfile_ffjb(filename);
	if (m_filetype == SFT_NONE)
		openfile_qt(filename);
	if (m_filetype == SFT_NONE) {
		fprintf(stderr, "could not play file.\n" );
		exit(1);
	}
}

jack_nframes_t dispatch_fillBuffer(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames )
{
	jack_nframes_t rv = 0;
	// TODO: phase invert ?! if (p_phaseinvert) ...
	switch (m_filetype) {
		case SFT_SNDFILE:
			rv=fillBuffer_sf(bufferptrs, frames );
			break;
		case SFT_QUICKTIME:
			rv=fillBuffer_qt(bufferptrs, frames );
			break;
		case SFT_FFMPEG:
			rv=fillBuffer_ffjb(bufferptrs, frames );
			break;
	}
	return (rv);
} 

void dispatch_sampleseek (jack_nframes_t sample) {
	switch (m_filetype) {
		case SFT_SNDFILE:
			sampleseek_sf(sample); 
			break;
		case SFT_QUICKTIME:
			sampleseek_qt(sample); 
			break;
		case SFT_FFMPEG:
			sampleseek_ffjb(sample); 
			break;
	}
}

///////////////////////////////////////////////////
// buffer resamled  

#if 0
jack_nframes_t fillBuffer_resample(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames );
jack_nframes_t fillcl_resample(audiocache *cache, jack_nframes_t startclock) {
	audiocache *cl;
	jack_nframes_t rv;
	
	cl= getcl(cache);
	cl->state&=~1; 
	rv=fillBuffer_resample(cl->d, CACHE_S);  // TODO set rv to 0 if frame is zero padded !
	if (rv > 0 ) { 
		cl->start=startclock;
		cl->end=startclock;
		cl->end+= rv;
		cl->state|=1;
//		printf("read line %i - %i!\n",cl->start,cl->end);
	}
	return (rv);
}

#endif

///////////////////////////////////////////////////
// resample 

extern jack_nframes_t j_bufsiz;
extern jack_nframes_t j_srate;

#include<samplerate.h>
#include<math.h>

jack_default_audio_sample_t **j_buffer = NULL;

void jadio_realloc_resample(jack_nframes_t frames ) {	
	int i;
	// TODO: allocate larger buffer for resampling 
	// (no realloc in RT callback!)
	if(0) printf("realloc buffer: %i frames\n",frames);
	if (j_buffer) {
		for (i=0;i<m_channels;i++) {
			free(j_buffer[i]);
		}
		free(j_buffer);
	}

	j_buffer=xcalloc(m_channels,sizeof(jack_default_audio_sample_t*));
	for (i=0;i<m_channels;i++) {
		j_buffer[i]=xcalloc((1.0+ceilf((float) frames/m_fResampleRatio)),sizeof(jack_default_audio_sample_t));
	}

	if (m_filetype==SFT_SNDFILE)
		realloc_sf((jack_nframes_t) (1.0+ceilf((float) frames/m_fResampleRatio)));
}

jack_nframes_t fillBuffer_resample(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames )
{

	SRC_DATA src_data;
	jack_nframes_t nread = (jack_nframes_t) ceilf((float)frames / m_fResampleRatio);
	jack_nframes_t rv =  dispatch_fillBuffer(j_buffer,nread+1);
	if (rv!= nread+1) return (0);

	// Fill all resampler parameter data...
	src_data.input_frames  = nread+1;
	src_data.output_frames = frames;
	src_data.end_of_input  = (nread < 1);
	src_data.src_ratio     =  m_fResampleRatio;
	src_data.input_frames_used = 0;
	src_data.output_frames_gen = 0;

	int i;
	for (i=0;i< m_channels; i++ ) {
		src_data.data_in       = j_buffer[i];
		src_data.data_out      = bufferptrs[i];
		//src_simple (&src_data, SRC_SINC_FASTEST, 1) ;
		src_simple (&src_data, SRC_SINC_MEDIUM_QUALITY, 1) ;
		//src_simple (&src_data, SRC_SINC_BEST_QUALITY, 1) ;
	}

	//	fprintf(stderr," %li %li (%li - %f) \n", src_data.output_frames_gen , src_data.input_frames_used, nread, m_fResampleRatio );

	return (frames);
} 

int joffset=0;

void sampleseek_resample (jack_nframes_t sample) {
	if (sample > joffset) sample-=joffset;//j_bufsiz; // seek ahead. 
	else sample=0;

	dispatch_sampleseek (sample);
}


///////////////////////////////////////////////////
// jadio - callback functions called by jack

/* 
 * void jadio_realloc(jack_nframes_t frames )
 *  called if jack changes it's buffersize 
 *
 * void sampleseek (jack_nframes_t sample)
 *  called once before each fillBuffer() indicating the start position 
 *  in jack sample-rate frames.
 *
 * jack_nframes_t fillBuffer(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames );
 *  read (frames) samples at current position, and store them in jack buffers
 *
 */

#if 1 // resample evey frame JIT

void initialize_decoder (void) { ; }

void jadio_realloc(jack_nframes_t frames ) {	
	jadio_realloc_resample(frames);
};

jack_nframes_t fillBuffer(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames ) {
	// FIXME: pre-alloc larger buffer - no alloc in RT callback.
	//jadio_realloc(frames);
	return fillBuffer_resample( bufferptrs,frames);
}

void sampleseek (jack_nframes_t sample) {
	sampleseek_resample(sample);
}


#else //// no resample ////

void initialize_decoder (void) { ; }

void jadio_realloc(jack_nframes_t frames ) {	
	if (m_filetype== SFT_SNDFILE)
		realloc_sf(frames);
}

jack_nframes_t fillBuffer(jack_default_audio_sample_t **bufferptrs, jack_nframes_t frames )
{
	return dispatch_fillBuffer(bufferptrs,frames);
}

void sampleseek (jack_nframes_t sample) {
	dispatch_sampleseek(sample);
}

#endif

#if 0
void mread (void) {
	char buf[16];
	int rx;
	if ((rx = fread(buf, sizeof(char),1,stdin)) > 0) {
		if (buf[0] == 'm') joffset+=100;
		if (buf[0] == 'n') joffset-=100;
	}
	printf("new offset: %i\n",joffset);
}
#include <sys/poll.h>
void loopthedebug(void) {
        int y=0;
        struct pollfd pfd;

        pfd.fd = 0;
        pfd.events = POLLIN | POLLPRI;
        while (1) {
       //        y=poll(&pfd,1,1000);
        //        if (y>0 && pfd.revents & (POLLIN | POLLPRI))
                  mread();
		printf("debugloop\n");
        }
}

void looptheCCCdebug(void) {
        fd_set fd;
        int max_fd=0;
        struct timeval tv = { 0, 0 };

	while (1) {
		tv.tv_sec = 1; tv.tv_usec = 0;

		FD_ZERO(&fd);
		FD_SET(0,&fd);
		max_fd=1;
		if (select(max_fd, &fd, NULL, NULL, &tv)) mread();
	}
}
#endif
